void acosf(void){}
void acoshf(void){}
void asinf(void){}
void asinhf(void){}
void atan2f(void){}
void atanf(void){}
void atanhf(void){}
void cabsf(void){}
void cacos(void){}
void cacosf(void){}
void cacosh(void){}
void cacoshf(void){}
void carg(void){}
void cargf(void){}
void casin(void){}
void casinf(void){}
void casinh(void){}
void casinhf(void){}
void catan(void){}
void catanf(void){}
void catanh(void){}
void catanhf(void){}
void cbrtf(void){}
void ccos(void){}
void ccosf(void){}
void ccosh(void){}
void ccoshf(void){}
void cexp(void){}
void cexpf(void){}
void clog(void){}
void clogf(void){}
void cosf(void){}
void coshf(void){}
void cpow(void){}
void cpowf(void){}
void cproj(void){}
void cprojf(void){}
void csin(void){}
void csinf(void){}
void csinh(void){}
void csinhf(void){}
void csqrt(void){}
void csqrtf(void){}
void ctan(void){}
void ctanf(void){}
void ctanh(void){}
void ctanhf(void){}
void erfcf(void){}
void erff(void){}
void exp2f(void){}
void expf(void){}
void expm1f(void){}
void hypotf(void){}
void lgammaf(void){}
void log10f(void){}
void log1pf(void){}
void log2f(void){}
void logf(void){}
void powf(void){}
void sinf(void){}
void sinhf(void){}
void sqrtf(void){}
void tanf(void){}
void tanhf(void){}
void tgammaf(void){}
